import logging
from gitlab import Gitlab
from gitlab.exceptions import GitlabCreateError, GitlabUpdateError, GitlabDeleteError
from datetime import datetime
from dateutil.relativedelta import relativedelta
from gitlab_usermgmt.user import User
from gitlab_usermgmt.access import Access

log = logging.getLogger(__name__)


class Api:
    """Interface to Gitlab API

    stores information, connectionm groups, projects, users
    """

    def __init__(self, group_id, dry_run):
        """
        API Init routine

        :param group_id: The source group id
        """
        """Gitlab connection object"""
        self.gl = None
        """Main group id"""
        self.group_id = int(group_id)
        """Subgroups under group_id"""
        self.subgroups = None
        """Projects under group_id"""
        self.projects = None
        """Gitlab Url"""
        self.url = None
        """Dry Run Identifier"""
        self.dry_run = dry_run

    def auth(self, token, url):
        """
        Authenticates against gitlab with your token
        :param token: The gitlab API token
        :param url: The gitlab URL, eg. https://gitlab.com
        :return:
        """
        log.debug("auth: ...")
        self.url = url
        self.gl = Gitlab(self.url, private_token=token)
        self.gl.auth()

    def get_users(self):
        """
        Logs in to Gitlab, queries all subgroups and projects,
        writes users.json file

        :param report_project: project where badges are linked against, eg. gitlab_usermgmt/people-ops
        :param access_definition_file: json dict of usergroups
        :return:
        """
        log.debug("write_user_access_list: ...")
        self.subgroups = self.get_all_subgroups(self.group_id)
        log.info("subgroups.len: " + str(self.subgroups.__len__()))
        self.projects = self.get_all_projects(self.subgroups)
        log.info("projects.len: " + str(self.projects.__len__()))
        users = get_all_users(self.subgroups, self.projects)
        log.info("users.len: " + str(users.__len__()))
        return users.copy()

    def get_all_subgroups(self, group_id):
        """
        Returns recursively all gitlab subgroup objects in a list of a source group
        """
        log.debug("get_all_subgroups: ...")
        group_list = []
        group = self.gl.groups.get(int(group_id))
        group_list.append(group)
        subgroups = group.subgroups.list(all=True)
        for subgroup in subgroups:
            log.debug(
                f"get_all_subgroups: append subgroup: {subgroup.id} - {subgroup.full_path}")
            group_list = group_list + self.get_all_subgroups(subgroup.id)
        return group_list

    def get_all_projects(self, subgroups):
        """
        Returns all gitlab project objects from a list of groups
        :param subgroups: list of gitlab group objects
        """
        log.debug("get_all_projects: ...")
        project_list = []
        for subgroup in subgroups:
            group_projects = subgroup.projects.list(all=True)
            for group_project in group_projects:
                log.debug(
                    f"get_all_projects: append group_projects: {group_project.id} - {group_project.path_with_namespace}")
                project_list.append(self.gl.projects.get(group_project.id))
        return project_list

    def _get_group(self, group_id):
        """
        Returns group with group_id from subgroups list

        :param group_id: The id for the group you want to have
        :return: A gitlab group object
        """
        for group in self.subgroups:
            if int(group.id) == int(group_id):
                return group
        raise ValueError(
            f"_get_group: Group with group_id: {group_id} not found.")

    def _get_project(self, project_id):
        """
        Returns project with project_id from subprojects list

        :param project_id: The id for the project you want to have
        :return: A gitlab project object
        """
        for project in self.projects:
            if project.id == int(project_id):
                return project
        raise ValueError(
            f"_get_project: project with project_id: {project_id} not found.")

    def _get_group_or_project(self, access):
        """
        Returns a group or project for an Access object

        :param access: Access Object
        """
        if access.access_type == "project":
            return self._get_project(access.id)
        elif access.access_type == "group":
            return self._get_group(access.id)
        else:
            raise ValueError(
                f"_get_group_or_project: Unsupported access_type: {access.access_type}")

    def remove(self, access, user_id):
        """
        Removes an access right an Gitlab

        :param access: Access object
        :param user_id: User ID
        """
        log.debug(
            f"remove: for user_id: {user_id} remove access to {access.access_type} {access.id} {access.web_url} to level {access.level} expiring at '{access.expires_at}'")
        user_id = int(user_id)
        try:
            accessobject = self._get_group_or_project(access)
            if not self.dry_run:
                accessobject.members.delete(user_id)
        except ValueError:
            log.warning(
                f'remove: Unable to remove {access} for user_id {user_id} to {access.access_type} with {access.web_url}')
        except GitlabDeleteError as e:
            if e.response_code == 403:
                log.warning(
                    "remove: 403 Member " + str(user_id) + " cannot be deleted in " + str(accessobject.id))
            elif e.response_code == 404:
                log.info(
                    "remove: 404 Member " + str(user_id) + " already deleted in " + str(accessobject.id))
            else:
                raise e

    def change(self, access, user_id):
        """
        Changes an access right an Gitlab

        :param access: Access object
        :param user_id: User ID
        """
        log.debug(
            f"change: for user_id: {user_id} change access to {access.access_type} {access.id} {access.web_url} to level {access.level} expiring at '{access.expires_at}'")
        user_id = int(user_id)
        try:
            accessobject = self._get_group_or_project(access)
            member = accessobject.members.get(user_id)
            member.access_level = access.level
            if access.expires_at != member.expires_at:
                member.expires_at = access.expires_at
            if not self.dry_run:
                member.save()
        except ValueError:
            log.warning(
                f'change: Unable to change access for user_id {user_id} to {access.access_type} with {access.web_url}')
        except GitlabUpdateError as e:
            if e.response_code == 403:
                log.warning(
                    "change: 403 Member " + str(user_id) + " cannot be modified in " + str(accessobject.id))
            elif e.response_code == 400:
                log.info(
                    "change: 400 Member " + str(user_id) + " has higher inherited access level to " + str(
                        accessobject.id))
            else:
                raise e

    def grant(self, access, user_id):
        """
        Grants an access right an Gitlab

        :param access: Access object
        :param user_id: User ID
        """
        log.debug(
            f"grant: for user_id: {user_id} grant access to {access.access_type} {access.id} {access.web_url} to level {access.level} expiring at '{access.expires_at}'")
        user_id = int(user_id)
        try:
            accessobject = self._get_group_or_project(access)
            if not self.dry_run:
                member = accessobject.members.create({
                    'user_id': user_id,
                    'access_level': access.level
                })
                member.save()
        except ValueError:
            log.warning(
                f'grant: Unable to grant access for user_id {user_id} to {access.access_type} with {access.web_url}')
        except GitlabCreateError as e:
            if e.response_code == 403:
                log.warning(
                    "grant: 403 Member " + str(user_id) + " cannot be added in " + str(accessobject.id))
            elif e.response_code == 409:
                log.info(
                    "grant: 409 Member " + str(user_id) + " already exists in " + str(accessobject.id))
            elif e.response_code == 400:
                log.info(
                    "grant: 400 Member " + str(user_id) + " has higher inherited access level to " + str(
                        accessobject.id))
            else:
                raise e

    def create_issue(self, project_id, assignee_id, issue_title, issue_text, issue_label_list, issue_due_date):
        """
        Creates or updates an issue with the title, text and label list
        Updates a issue if one is found with the same title.
        Adds timestamp to issue text to trigger email notifications when issue text stays the same

        :param project_id: ID of the project where the issue will be created
        :param assignee_id: Assigne User ID
        :param issue_title: Title of the issue
        :param issue_text: Content of the issue description
        :param issue_label_list: Labels that will be attached to the issue
        :param issue_due_date: Due date string %Y-%m-%d of the issue
        """
        log.debug(
            f"create_issue: {issue_title} issue_label_list: {issue_label_list} issue_due_date: {issue_due_date}")
        timestamp_footer = '<br><br>' + str(datetime.now())
        # get the project id
        project = self._get_project(int(project_id))
        # get all issues with label `expire`
        issues = project.issues.list(labels=issue_label_list, all=True)
        existing_issue = None
        # find first issue where the title match
        for issue in issues:
            if issue.attributes["title"] == issue_title:
                existing_issue = issue
                break
        # update existing issue
        if existing_issue is not None:
            log.debug(
                f"create_issue: {issue_title} - update existing issue: {existing_issue.web_url} created at: {existing_issue.created_at} updated at: {existing_issue.updated_at}")
            # only update an issue once per week to limit mail spam, updated_at format: '2020-05-22T07:19:15.482Z'
            created_date = datetime.strptime(
                existing_issue.updated_at.split("T")[0], "%Y-%m-%d")
            if created_date < (datetime.now() - relativedelta(days=6)):
                existing_issue.due_date = issue_due_date
                existing_issue.description = issue_text + timestamp_footer
                existing_issue.state_event = 'reopen'
                if not self.dry_run:
                    existing_issue.save()
            else:
                log.info(
                    f"create_issue: {issue_title} - skip update existing issue because last update < 6 days ago: {existing_issue.web_url} created at: {existing_issue.created_at} updated at: {existing_issue.updated_at}")
        else:
            log.debug(f"create_issue: {issue_title} - create new issue")
            if not self.dry_run:
                issue = project.issues.create({
                    "title": issue_title,
                    "description": issue_text + timestamp_footer,
                    "assignee_ids": int(assignee_id),
                    "labels": issue_label_list,
                    "due_date": issue_due_date
                })


def add_user_to_dict(user_dict, user_id, user_name, user_state, access_id, access_type, access_level, access_expire, access_weburl):
    user_id = int(user_id)
    access_id = int(access_id)
    if user_id not in user_dict:
        user = User(user_id, user_name, user_state)
    else:
        user = user_dict[user_id]
    access = Access(access_id, access_type, access_level,
                    access_expire, access_weburl)
    user.set_user_access(access)
    user_dict[user_id] = user


def get_all_users(subgroups, projects):
    """
    Returns a dict of users with their user access to groups and projects
    :param subgroups: list of gitlab group objects
    :param projects:  list of gitlab project objects
    :return: dict of users
    """
    log.debug("get_all_users: ...")
    user_dict = {}
    # get all users from groups first
    for subgroup in subgroups:
        group_members = subgroup.members.list(all=True)
        for group_member in group_members:
            add_user_to_dict(user_dict,
                             group_member.id,
                             group_member.username,
                             group_member.state,
                             subgroup.id,
                             "group",
                             group_member.access_level,
                             group_member.expires_at,
                             subgroup.web_url)
    # get all users from the projects
    for project in projects:
        project_members = project.members.list(all=True)
        for project_member in project_members:
            add_user_to_dict(user_dict,
                             project_member.id,
                             project_member.username,
                             project_member.state,
                             project.id,
                             "project",
                             project_member.access_level,
                             project_member.expires_at,
                             project.web_url)
    return user_dict
