from datetime import datetime
from dateutil.relativedelta import relativedelta
from jinja2 import Template
import logging

log = logging.getLogger(__name__)


class User:
    """
    User Object

    """

    def __init__(self, id, username, state):
        """
        Initializes the user object

        :param id: Gitlab User ID
        :param username: The Gitlab username
        :param state: The Gitlab user state identifier
        """
        self.id = int(id)
        self.username = username
        self.state = state
        """ Expire date string """
        self.expires_at = ""
        """ Meta information dict of the user from config """
        self.meta = {}
        """ Access object dicts that define group access as it is currently in GitLab """
        self.group_access = {}
        """ Access object dicts that define project access as it is currently in GitLab """
        self.project_access = {}
        """ access_list defines access objects from config.json """
        self.access_list = []
        """ to_be dicts contain access objects which where generated by compare_access """
        self.group_to_be_removed = {}
        self.group_to_be_changed = {}
        self.group_to_be_granted = {}
        self.project_to_be_removed = {}
        self.project_to_be_changed = {}
        self.project_to_be_granted = {}

    def set_user_access(self, access):
        """
        Sets access as it is in Gitlab to in self.group_access or self.project_access

        :param access: Access object
        """
        log.debug(
            f"set_user_access: user id: {self.id} access id: {access.id} access_type: {access.access_type}")

        if access.access_type == "group":
            self.group_access[access.id] = access
        elif access.access_type == "project":
            self.project_access[access.id] = access
        else:
            raise ValueError(
                f"User: set_user_access: Unsupported access_type: {access.access_type} for user {self.username}")

    def find_higher_or_equal_inherited_access(self, access, gitlab_url="https://gitlab.com"):
        """
        Returns true if higher or equal inherited access exists in defined access_list for a given access.web_url.

        :param access: Access object that we will search for
        :param gitlab_url: The gitlab web url
        :returns: True when higher or equal access is found, otherwise False
        """
        group_access_list = self.get_usergroups_for_path(
            access.web_url, True, gitlab_url)
        for group_access_id, group_access_item in group_access_list.items():
            if group_access_id != access.id:
                if group_access_item.level >= access.level:
                    return True
        return False

    def find_any_inherited_access(self, access, gitlab_url="https://gitlab.com"):
        """
        Returns true if any inherited access exists in defined access_list for a given access.web_url.

        :param access: Access object that we will search for
        :param gitlab_url: The gitlab web url
        :returns: True when higher or equal access is found, otherwise False
        """
        group_access_list = self.get_usergroups_for_path(
            access.web_url, True, gitlab_url)
        for group_access_id, group_access_item in group_access_list.items():
            if group_access_id != access.id:
                return True
        return False

    def compare_access(self, set_expire_dates, set_access_level, gitlab_url="https://gitlab.com"):
        """
        Compares user defined access_list with existing access dicts self.group_access and self.project_access
        Sets group/project_to_be_removed/changed/granted

        :param set_expire_dates: If true it will also compare the expire date, if false it will ignore it
        :param set_access_level: If true it will also compare the access level, if false it will ignore it
        :param gitlab_url: url for checking for inherited access
        """
        log.debug(
            f"compare_access: user id: {self.id} set_expire_dates: {set_expire_dates} set_access_level: {set_access_level}")
        # lists for ids that will be compared to existing access
        project_ids = []
        group_ids = []
        # iterate over all access items
        for access in self.access_list:
            # decide if we have a group or project
            if access.access_type == "project":
                project_ids.append(access.id)
                # check if we already have access
                if access.id in self.project_access:
                    # check if access level or expire date differs
                    if (set_access_level and int(access.level) != self.project_access[access.id].level) or (set_expire_dates and access.expires_at != self.project_access[access.id].expires_at):
                        # add access to to_be_changed list
                        log.info(
                            f"compare_access: user id: {self.id} change access - project id: {access.id} old level: {self.project_access[access.id].level} new level: {access.level} - old expire: {self.project_access[access.id].expires_at} new expire: {access.expires_at}")
                        if self.find_higher_or_equal_inherited_access(access, gitlab_url):
                            log.info(
                                f"compare_access: user id: {self.id} change access skipped because inherited access - project id: {access.id} {access.web_url}")
                        else:
                            self.project_to_be_changed[access.id] = access
                else:
                    # we do not have access, add to grant list
                    log.info(
                        f"compare_access: user id: {self.id} grant access - project id: {access.id} new level: {access.level} - new expire: {access.expires_at}")
                    if self.find_higher_or_equal_inherited_access(access, gitlab_url):
                        log.info(
                            f"compare_access: user id: {self.id} grant access skipped because inherited access - project id: {access.id} {access.web_url}")
                    else:
                        self.project_to_be_granted[access.id] = access
            elif access.access_type == "group":
                group_ids.append(access.id)
                # check if we already have access
                if access.id in self.group_access:
                    # check if access level differs
                    if (set_access_level and int(access.level) != self.group_access[access.id].level) or (set_expire_dates and access.expires_at != self.group_access[access.id].expires_at):
                        # add access to to_be_changed list
                        log.info(
                            f"compare_access: user id: {self.id} change access - group id: {access.id} old level: {self.group_access[access.id].level} new level: {access.level} - old expire: {self.group_access[access.id].expires_at} new expire: {access.expires_at}")
                        if self.find_higher_or_equal_inherited_access(access, gitlab_url):
                            log.info(
                                f"compare_access: user id: {self.id} change access skipped because inherited access - group id: {access.id} {access.web_url}")
                        else:
                            self.group_to_be_changed[access.id] = access
                else:
                    # we do not have access, add to grant list
                    log.info(
                        f"compare_access: user id: {self.id} grant access - group id: {access.id} new level: {access.level} - new expire: {access.expires_at}")
                    if self.find_higher_or_equal_inherited_access(access, gitlab_url):
                        log.info(
                            f"compare_access: user id: {self.id} grant access skipped because inherited access - group id: {access.id} {access.web_url}")
                    else:
                        self.group_to_be_granted[access.id] = access
            else:
                raise ValueError(
                    f"compare_access: unsupported access_type {access.access_type} for user id: {self.id}")
        # remove existing access that was not defined in self.access_list
        for access_id, access_item in self.project_access.items():
            if access_id not in project_ids:
                if not set_access_level and self.find_any_inherited_access(access_item, gitlab_url):
                    log.info(
                        f"compare_access: user id: {self.id} remove access skipped because inherited access - project id: {access_id} {access_item.web_url}")
                else:
                    log.info(
                        f"compare_access: user id: {self.id} remove access to project {access_id} {access_item.web_url}")
                    self.project_to_be_removed[access_id] = access_item
        for access_id, access_item in self.group_access.items():
            if access_id not in group_ids:
                if not set_access_level and self.find_any_inherited_access(access_item, gitlab_url):
                    log.info(
                        f"compare_access: user id: {self.id} remove access skipped because inherited access - project id: {access_id} {access_item.web_url}")
                else:
                    log.info(
                        f"compare_access: user id: {self.id} remove access to group {access_id} {access_item.web_url}")
                    self.group_to_be_removed[access_id] = access_item
        log.debug(f"compare_access: user id: {self.id} finished.")

    def manage_gitlab_access(self, api):
        """
        Manages users compared access states in Gitlab

        :param api: A gitlab_usermgmt API object
        """
        # TODO: what happens when we set an expire date to the past?
        log.debug(f"manage_gitlab_access: user id: {self.id}")
        if (len(self.group_to_be_removed) + len(self.group_to_be_changed) + len(self.group_to_be_granted) + len(self.project_to_be_removed) + len(self.project_to_be_changed) + len(self.project_to_be_granted)) == 0:
            log.debug(
                f"manage_gitlab_access: Nothing to do for user id: {self.id}")
            return

        for item_id, item in self.group_to_be_removed.items():
            log.debug(
                f"manage_gitlab_access: remove group access for user id: {self.id} item_id: {item_id}")
            api.remove(item, self.id)
        for item_id, item in self.group_to_be_changed.items():
            log.debug(
                f"manage_gitlab_access: change group access for user id: {self.id} item_id: {item_id}")
            api.change(item, self.id)
        for item_id, item in self.group_to_be_granted.items():
            log.debug(
                f"manage_gitlab_access: grant group access for user id: {self.id} item_id: {item_id}")
            api.grant(item, self.id)
        for item_id, item in self.project_to_be_removed.items():
            log.debug(
                f"manage_gitlab_access: remove project access for user id: {self.id} item_id: {item_id}")
            api.remove(item, self.id)
        for item_id, item in self.project_to_be_changed.items():
            log.debug(
                f"manage_gitlab_access: change project access for user id: {self.id} item_id: {item_id}")
            api.change(item, self.id)
        for item_id, item in self.project_to_be_granted.items():
            log.debug(
                f"manage_gitlab_access: grant project access for user id: {self.id} item_id: {item_id}")
            api.grant(item, self.id)

    def _create_expire_issue_text(self, template_text):
        """
        Returns expire issue text

        :param template_text: The Jinja template
        :returns expire issue text:
        """
        log.debug(f"_create_expire_issue_test: user id: {self.id}")
        t = Template(template_text)

        # render issue text and replace line breaks with nothing
        issue_text = t.render(
            user=self,
        ).replace('\n', '')
        return issue_text

    def create_expire_issue(self, api, report_project, template_text, expire_in_days):
        """
        Creates expire issues for this user in Gitlab

        :param api: A gitlab_usermgmt API object
        :param report_project: ID of the project where the isssues will be created
        :param template_text: The Jinja template
        :param expire_in_days: Creates expire issue if expring in less eq than specified days
        """
        log.debug(
            f"create_expire_issue: user id: {self.id} with expire date: {self.expires_at}")
        # check if user state is not active and return directly
        if self.state != "active":
            log.warning(
                f"create_expire_issue: cannot create expire issue for user id: {self.id} because user is not active anymore.")
        # check if expire date is in less than 30 days
        if self.expires_at == "":
            log.warning(
                f"create_expire_issue: cannot create expire issue for user id: {self.id} because no date is defined.")
            return
        # check if expire date is longer than 1 year in future
        expire_date_max = datetime.now() + relativedelta(years=1)
        expire_date = datetime.strptime(self.expires_at, "%Y-%m-%d")
        if expire_date_max < expire_date:
            log.warning(
                f"create_expire_issue: expire date for user id: {self.id} is more than one year in the future: {self.expires_at}")
            return
        # check if expire date is in the past
        if datetime.now() > expire_date:
            log.warning(
                f"create_expire_issue: expire date for user id: {self.id} is in the past: {self.expires_at} Remove him from access definition list!")
            return
        # check if expire date is in less than 30 days
        expire_date_due = datetime.now() + relativedelta(days=expire_in_days)
        if expire_date_due >= expire_date:
            # create expire issue
            issue_text = self._create_expire_issue_text(template_text)
            issue_title = f"Expiring access {self.username}"
            issue_label_list = ["expire"]
            issue_due_date = self.expires_at
            api.create_issue(report_project, self.id, issue_title,
                             issue_text, issue_label_list, issue_due_date)

    def get_usergroups_for_path(self, path, return_inherited, gitlab_url="https://gitlab.com"):
        """
        Returns a dict of access objects for a namespace path, when access objects with that path
        where found in access_list, does not check for actual set access in project_access and group_access
        because we want to set it as it is defined in access_list

        :param path: the path_with_namespace attribute
        :param return_inherited: if true do return inherited access definitions, checks if url starts with path then
        :param gitlab_url: is used to remove prefix, default: https://gitlab.com
        """
        log.debug("get_usergroups_for_path: path: " + str(path))
        access_found_dict = dict()  # dict of access objects
        path = path.replace(gitlab_url + "/", "")
        if path.startswith("groups/"):
            path = path.replace("groups/", "")
        if path.endswith("/"):
            path = path[:-1]
        # iterate over access definition items
        for access in self.access_list:
            found = False
            if access.access_type == "group":
                # remove gitlab domain name
                url = access.web_url.replace(gitlab_url + "/", "")
                # remove groups as well
                if url.startswith("groups/"):
                    url = url.replace("groups/", "")
                if path == url:  # check if path is exactly the url
                    found = True
                # check if path starts with entry url
                elif return_inherited:
                    path_split = path.split("/")
                    access_url_split = url.split("/")
                    min_elements = min(len(path_split), len(access_url_split))
                    for i in range(0, min_elements):
                        if path_split[i] == access_url_split[i]:
                            if (i + 1) == min_elements:
                                found = True
                if found:
                    # if higher access level already exists, continue
                    if access_found_dict.__contains__(access.id):
                        if access.level < access_found_dict[access.id].level:
                            continue
                    access_found_dict[access.id] = access
            elif access.access_type == "project":
                # remove gitlab domain name
                url = access.web_url.replace(gitlab_url + "/", "")
                if path == url:  # check if path is exactly the url
                    found = True
                if found:
                    # if higher access level already exists, continue
                    if access_found_dict.__contains__(access.id):
                        if access.level < access_found_dict[access.id].level:
                            continue
                    access_found_dict[access.id] = access
            else:
                log.warning(
                    f"get_usergroups_for_path: unsupported access type {access.access_type} for user {self.id}")
        return access_found_dict
