from gitlab_usermgmt.api import Api
from gitlab_usermgmt.io import write_dict_to_json
from gitlab_usermgmt.access import Access
from gitlab_usermgmt.user import User
from datetime import datetime
import os
import logging

log = logging.getLogger(__name__)


class Usermgmt:
    """
    Main Class for GitLab Usermanagement
    """

    def __init__(self, config_dict, dry_run):
        """
        Initializes class objects, sets the configuration parameters
        """
        log.debug("__init__: ...")
        self.gitlab_url = config_dict["settings"]["gitlab_url"]
        self.group_id = int(config_dict["settings"]["group_id"])
        self.report_project = config_dict["settings"]["report_project"]
        self.dry_run = dry_run
        self.set_expire_dates = config_dict["settings"]["actions"]["set_expire_dates"]
        self.set_access_level = config_dict["settings"]["actions"]["set_access_level"]
        self.create_expire_issues = config_dict["settings"]["actions"]["create_expire_issues"]
        self.expire_in_days = config_dict["settings"]["expire_in_days"]
        self.expire_issue_template = config_dict["settings"]["expire_issue_template"]
        self.access_definition = config_dict["access_definition"]
        """GitLab API connection object"""
        self.api = Api(self.group_id, dry_run)
        """Users under group_id"""
        self.users = {}
        log.debug("__init__: done.")

    def auth(self):
        """
        Authenticates at GitLab with environment token
        """
        log.debug("auth: starting ...")
        token = os.getenv('GITLAB_TOKEN')
        if token is None:
            raise (ValueError(
                'GITLAB_TOKEN environment variable is None. Set it with your GitLab API Token.'))
        if token == "":
            raise (ValueError(
                'GITLAB_TOKEN environment variable is empty. Set it with your GitLab API Token.'))
        self.api.auth(token, self.gitlab_url)
        log.debug("auth: done.")

    def _remove_duplicates_from_access_list(self, access_list):
        """
        Removes duplicates from access list
         sets only the highest access_level that is found
        """
        log.debug("_remove_duplicates_from_access_list: ...")
        projects = {}
        groups = {}
        result = []
        for access in access_list:
            if access.access_type == "project":
                # find access in projects
                if access.id in projects:
                    # compare access level
                    if projects[access.id].level < access.level:
                        projects[access.id] = access
                else:
                    # not found, add to list
                    projects[access.id] = access
            if access.access_type == "group":
                # find access in groups
                if access.id in groups:
                    # compare access level
                    if groups[access.id].level < access.level:
                        groups[access.id] = access
                else:
                    # not found, add to list
                    groups[access.id] = access
        # make plain list out of dicts
        for id, access in projects.items():
            result.append(access)
        for id, access in groups.items():
            result.append(access)
        log.debug("_remove_duplicates_from_access_list: done.")
        return result

    def _parse_expires_at(self, user_info):
        """
        Returns expires_at info from a users access definition info

        :returns expires_at string
        """
        expires_at = ""
        if "meta" in user_info:
            if "expires_at" in user_info["meta"]:
                if user_info["meta"]["expires_at"] != "":
                    try:
                        datetime.strptime(
                            user_info["meta"]["expires_at"], "%Y-%m-%d")
                        expires_at = user_info["meta"]["expires_at"]
                    except ValueError as e:
                        log.warning(
                            f"_parse_expires_at: expires_at parsing error: {e.args[0]} for username: {user_info['username']} is empty!")
                else:
                    log.warning(
                        f"_parse_expires_at: expires_at entry in meta dict for username: {user_info['username']} is empty!")
            else:
                log.warning(
                    f"_parse_expires_at: missing expires_at entry in meta dict for username: {user_info['username']}!")
        else:
            log.warning(
                f"_parse_expires_at: missing meta dict entry for username: {user_info['username']}!")
        return expires_at

    def _set_access_definitions_for_user(self, user_id):
        """
        sets list of access definitions for a specific user id in users dict

        :param user_id: Gitlab User ID
        :returns: list of access definitions for that user id
        """
        log.debug("_set_access_definitions_for_user: ...")
        access_list = []
        expires_at = ""
        user_id_str = str(user_id)
        # check for str of id in access definition because they are string keys there
        if user_id_str in self.access_definition["users"]:
            if "meta" in self.access_definition["users"][user_id_str]:
                self.users[user_id].meta = self.access_definition["users"][user_id_str]["meta"]
            # get expires_at data if expire dates are enabled
            if self.set_expire_dates:
                expires_at = self._parse_expires_at(
                    self.access_definition["users"][user_id_str])
            # walk over all usergroups for this user
            for usergroupname in self.access_definition["users"][user_id_str]["usergroups"]:
                # add access items for that usergroup to access list
                access_dict = self.access_definition["usergroups"][usergroupname]["access"]
                for entry in access_dict:
                    access = Access(entry["id"], entry["access_type"],
                                    entry["access_level"], expires_at, entry["web_url"])
                    access_list.append(access)
        else:
            log.warning(
                f"_set_access_definitions_for_user: User ID {user_id} not found in access_definition")
        self.users[user_id].access_list = self._remove_duplicates_from_access_list(
            access_list)
        self.users[user_id].expires_at = expires_at
        log.debug("_get_access_definitions_for_user: done.")

    def _compare_user_access(self):
        """
        Compares defined user access rights to those from Gitlab
        """
        log.debug("_compare_user_access: ...")
        # add users that are defined but do not have access yet
        for user_id, user_info in self.access_definition["users"].items():
            user_id = int(user_id)
            if user_id not in self.users:
                log.info(
                    f"_compare_user_access: user_id: {user_id} not found in existing in Gitlab users list")
                self.users[user_id] = User(
                    user_id, user_info["username"], "active")
        # add access definitions and compare access for all users
        for user_id, user_info in self.users.items():
            log.debug("_compare_user_access: user_id: " + str(user_id))
            self._set_access_definitions_for_user(user_id)
            user_info.compare_access(
                self.set_expire_dates, self.set_access_level)
        log.debug("_compare_user_access: done.")

    def _manage_gitlab_access(self):
        """
        Sets access rights at Gitlab
        """
        log.debug("_manage_gitlab_access: ...")
        for user_id, user_info in self.users.items():
            log.debug("_manage_gitlab_access: user_id: " + str(user_id))
            user_info.manage_gitlab_access(self.api)
        log.debug("_manage_gitlab_access: done.")

    def _create_expire_issues(self):
        """
        Creates expire issues for users in Gitlab
        """
        log.debug("_create_expire_issues: ...")
        for user_id, user_info in self.users.items():
            log.debug("_create_expire_issues: user_id: " + str(user_id))
            f = open(self.expire_issue_template, "r")
            template_text = f.read()
            user_info.create_expire_issue(
                self.api, self.report_project, template_text, self.expire_in_days)
        log.debug("_create_expire_issues: done.")

    def run(self):
        """
        Executes usermgmt as defined by init
        """
        log.debug("run: starting ...")

        self.users = self.api.get_users()
        write_dict_to_json(self.users, "users.json")

        self._compare_user_access()
        write_dict_to_json(self.users, "users_compared.json")

        self._manage_gitlab_access()

        if self.create_expire_issues:
            self._create_expire_issues()

        log.debug("run: finished.")
