import json
import logging

log = logging.getLogger(__name__)


def validate_access_definition_file(filename):
    """
    Validates a access definition file

    :param filename: reads the json file into a dict
    :returns: true if valid, throws logger warnings and errors otherwise
    """
    log.debug("validate_access_definition_file: filename: " + filename)
    f = open(filename, "r")
    data = json.loads(f.read())
    return validate_access_definition(data)


def validate_access_definition(data):
    """
    Validates a access definition dictionary

    :param data: access definition dictionary
    :returns: true if valid, throws logger warnings or errors otherwise
    """
    log.debug("validate_access_definition: data: " + str(data))
    valid = True
    # check if data is empty
    if data == "":
        raise (ValueError('The file is empty.'))
    # check if usergroups are missing
    if "usergroups" not in data:
        log.error('usergroups are missing.')
        valid = False
    # check if users are missing
    if "users" not in data:
        log.error('users are missing.')
        valid = False
    # only do additional checks when basic structure is ok
    if valid:
        # validate usergroup structure
        for group_id, group_info in data["usergroups"].items():
            # check if approvers exist
            if "meta" not in group_info:
                log.warning("meta entry missing for usergroup with id: " + str(group_id))
            else:
                if "approvers" not in group_info["meta"]:
                    log.warning("approvers entry missing for usergroup with id: " + str(group_id))
                else:
                    # check if all approvers exist in users ist
                    for approver in group_info["meta"]["approvers"]:
                        # try to find approver in users list
                        approver_found = False
                        for user_id, user_info in data["users"].items():
                            if user_info["username"] == approver:
                                approver_found = True
                        if not approver_found:
                            log.error(
                                "The approver " + approver + " for usergroup " + group_id +
                                " can not be found in users list.")
                            valid = False
            # check structure of project or group entries in the usergroup
            for a_id, a_info in group_info.items():
                if str(a_id) == "meta":
                    continue
                if "web_url" not in a_info:
                    log.error('web_url is missing for entry ' + str(a_id) + " in usergroup " + str(group_id))
                    valid = False
                if "access_level" not in a_info:
                    log.error('web_url is missing for entry ' + str(a_id) + " in usergroup " + str(group_id))
                    valid = False
        # check if all assigned usergroups under are also defined
        for user_id, user_info in data["users"].items():
            for group in user_info["usergroups"]:
                # find usergroup in usergroups
                if group not in data["usergroups"]:
                    log.error("Usergroup " + group + " is defined for user " + str(user_id) +
                              " - " + user_info["username"] + " but can not be found in usergroup list.")
                    valid = False

    # raise error if config is not valid
    if not valid:
        raise (ValueError('access_definition is not valid! Check the logs for errors and warnings!'))
    return True
